/*
 * Copyright (c) 2013-2014, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * \file ti_em_init.c
 *
 * This file implements additional functions, types, variables and constants to 
 * be used for initializing the EM
 * The implementations are TI specific and are likely the same for all 
 * applications
 */

#include <stdio.h>
#include "ti_em_init.h"
#include "ocl_em_init.h"
#include "util.h"
#include "monitor.h"
#include <ti/runtime/openem/firmware/event_machine_firmware.h>
#ifdef TI_EM_C6678
#include <ti/runtime/openem/src/event_machine_private_macros.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#endif
#include <ti/drv/cppi/cppi_desc.h>
#include <ti/csl/csl_cppi.h>

extern cregister volatile uint32_t DNUM;

#define Emti_REGION_NUM (3)
#define Emti_PUBLIC_REGION_IDX (0)
#define Emti_PRIVATE0_REGION_IDX (1)
#define Emti_PRIVATE1_REGION_IDX (2)

#define Emti_PDSP_NUM (1)
#define Emti_PDSP_IDX (0)

#ifdef TI_EM_C6678
// #define OCLPDSP 0x340B8000  // old addresses
// #define OCKPDSP 0x340BC000  // will use direct aliased address for PDSP
// #define OCLPDSP 0x02AB8000     // these aliased address work on shannon
// #define OCLPDSP 0x02ABC000
#define OCLPDSP 0x02ABC000
#endif

#ifdef TI_EM_C6634
// #define OCLPDSP 0x23A1E000      // PDSPSH works! 
// #define OCLPDSP 0x23A20000   // PDSP1D 
// #define OCLPDSP 0x23A24000   // PDSP2D works
// #define OCLPDSP 0x23A28000   // PDSP3D
// #define OCLPDSP 0x23A2C000   // PDSP4D
#define OCLPDSP 0x23A30000   // PDSP5D works
#endif


#ifndef TI_EM_C6634
extern Qmss_GlobalConfigParams qmssGblCfgParams;
extern Cppi_GlobalConfigParams cppiGblCfgParams[];
#else
#include <ti/drv/cppi/device/k2k/src/cppi_device.c>
#include <ti/drv/qmss/device/k2k/src/qmss_device.c>
#endif

FAST_SHARED_1D(uint8_t, Emti_svPublicEventDscMem, TI_EM_EVENT_NUM2 * TI_EM_PUBLIC_EVENT_DSC_SIZE);


uint32_t
Emti_roundDscNum (
    uint32_t eventNum
)
{
    uint32_t lvEventNum;
    uint32_t lvEventNumTmp;
    uint32_t lvEventNumMin = 0x00000020;    // 32
    uint32_t lvEventNumMax = 0x00004000;    // 16K

    lvEventNum = eventNum;
    if (lvEventNum < lvEventNumMin)
    {
        lvEventNum = lvEventNumMin;
    }
    else
    {
        lvEventNumTmp = lvEventNumMax >> 1;
        while (lvEventNum <= lvEventNumTmp)
        {
            lvEventNumTmp = lvEventNumTmp >> 1;
        }
        lvEventNum = lvEventNumTmp << 1;
    }

    return lvEventNum;
}

uint32_t
Emti_makeAddressGlobal (
    uint32_t coreIdx,
    uint32_t address
)
{
    if (address < 0x01000000)
    {
        address = (0x10 + coreIdx) << 24 | address;
    }

    return address;
}

int
Emti_initPdspConfig (
    uint32_t pdspNum1,
    Qmss_PdspCfg * pdspConfigTbl1,
    Qmss_PdspCfg * pdspConfigTbl2,
    Qmss_PdspId * emPdspIdxPtr
)
{
    uint32_t lvPdspNum;

#ifdef DSP_EXAMPLE_STANDALONE
    lvPdspNum = pdspNum1 + Emti_PDSP_NUM;
#else
    lvPdspNum = pdspNum1 + 1;
#endif
    if (lvPdspNum > QMSS_MAX_PDSP)
        return -1;

#ifdef TI_EM_C6670
    *emPdspIdxPtr = Qmss_PdspId_PDSP2;
    pdspConfigTbl2[Emti_PDSP_IDX].pdspId = Qmss_PdspId_PDSP2;
    pdspConfigTbl2[Emti_PDSP_IDX].firmware = &em_scheduler_router_fw_c667x_pdsp2;
    pdspConfigTbl2[Emti_PDSP_IDX].size = sizeof (em_scheduler_router_fw_c667x_pdsp2);
#endif
#ifdef TI_EM_C6678
    *emPdspIdxPtr = Qmss_PdspId_PDSP2;
    pdspConfigTbl2[Emti_PDSP_IDX].pdspId = Qmss_PdspId_PDSP2;
    pdspConfigTbl2[Emti_PDSP_IDX].firmware = &em_scheduler_router_fw_c667x_pdsp2;
    pdspConfigTbl2[Emti_PDSP_IDX].size = sizeof (em_scheduler_router_fw_c667x_pdsp2);
#endif
#ifdef TI_EM_C6634
    *emPdspIdxPtr = Qmss_PdspId_PDSP5;
    pdspConfigTbl2[Emti_PDSP_IDX].pdspId = Qmss_PdspId_PDSP5;
    pdspConfigTbl2[Emti_PDSP_IDX].firmware = &em_scheduler_router_fw_c663x;
    pdspConfigTbl2[Emti_PDSP_IDX].size = sizeof (em_scheduler_router_fw_c663x);
#endif
/***
// YUAN: we are only using one PDSP, no need to route to ARM EM
#ifdef TI_EM_C6634
    pdspConfigTbl2[Emti_PDSP_IDX + 1].pdspId = Qmss_PdspId_PDSP3;
    pdspConfigTbl2[Emti_PDSP_IDX + 1].firmware = &em_router_fw_c6634;
    pdspConfigTbl2[Emti_PDSP_IDX + 1].size = sizeof (em_router_fw_c6634);
#endif
***/
    return lvPdspNum;
}

// At least 1 region for public global events
// Exactly 1 region for private events
// At least 1 region per core for public local events
int
Emti_initRegionConfig (
    uint32_t regionNum1,
    Qmss_MemRegInfo * regionConfigTbl1,
    Qmss_MemRegInfo * regionConfigTbl2
)
{
    uint32_t lvRegionNum;
    uint32_t lvDscNum = 0;
    uint32_t lvLinkIdx = MY_DSP_EM_FIRST_DESC_IDX_IN_LINKING_RAM;
    Qmss_MemRegInfo *lvRegionConfigPtrTbl[QMSS_MAX_MEM_REGIONS];
    int i, j;

    lvRegionNum = regionNum1 + Emti_REGION_NUM;
    if (lvRegionNum > QMSS_MAX_MEM_REGIONS)
        return -1;

    // initialize descriptor regions for OpenEM
    // public event descriptors
    regionConfigTbl2[Emti_PUBLIC_REGION_IDX].descSize = (uint32_t) TI_EM_PUBLIC_EVENT_DSC_SIZE;
    regionConfigTbl2[Emti_PUBLIC_REGION_IDX].descNum = Emti_roundDscNum (TI_EM_EVENT_NUM2);
    regionConfigTbl2[Emti_PUBLIC_REGION_IDX].descBase =
        (uint32_t *) Emti_makeAddressGlobal (0, (uint32_t) Emti_svPublicEventDscMem);
    regionConfigTbl2[Emti_PUBLIC_REGION_IDX].manageDescFlag = Qmss_ManageDesc_UNMANAGED_DESCRIPTOR;

    // private event descriptors (CD region)
    regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descSize = (uint32_t) TI_EM_PRIVATE_EVENT_DSC_SIZE;
    regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descNum = Emti_roundDscNum ((uint32_t) TI_EM_CD_PRIVATE_EVENT_NUM);

    // TODO: do we need differentiation?
    regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descBase = (uint32_t *) 
                                  Emti_makeAddressGlobal (0, (uint32_t) OCLPDSP);
    memset((void*)OCLPDSP, 0, 0x4000);
    disableCache(OCLPDSP >> 24, OCLPDSP >> 24);

    regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].manageDescFlag = Qmss_ManageDesc_UNMANAGED_DESCRIPTOR;

    // private event descriptors (AP region)
    regionConfigTbl2[Emti_PRIVATE1_REGION_IDX].descSize = (uint32_t) TI_EM_PRIVATE_EVENT_DSC_SIZE;
    regionConfigTbl2[Emti_PRIVATE1_REGION_IDX].descNum = Emti_roundDscNum (TI_EM_AP_PRIVATE_EVENT_NUM);
    regionConfigTbl2[Emti_PRIVATE1_REGION_IDX].descBase =
        (uint32_t *) ((uint32_t) regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descBase +
                      regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descNum *
                      regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descSize);
    regionConfigTbl2[Emti_PRIVATE1_REGION_IDX].manageDescFlag = Qmss_ManageDesc_UNMANAGED_DESCRIPTOR;

    // copy user descriptor regions
    for (i = 0; i < regionNum1; i++)
    {
        regionConfigTbl2[Emti_REGION_NUM + i].descSize = regionConfigTbl1[i].descSize;
        regionConfigTbl2[Emti_REGION_NUM + i].descNum = regionConfigTbl1[i].descNum;
        regionConfigTbl2[Emti_REGION_NUM + i].descBase = regionConfigTbl1[i].descBase;
        regionConfigTbl2[Emti_REGION_NUM + i].manageDescFlag = regionConfigTbl1[i].manageDescFlag;
    }

    // Sort descriptor regions according to memory base address.
    // Count total number of descriptors.
    for (i = 0; i < lvRegionNum; i++)
    {
        uint32_t lvRank = 0;
        for (j = 0; j < lvRegionNum; j++)
        {
            if (regionConfigTbl2[j].descBase < regionConfigTbl2[i].descBase)
                lvRank++;
        }
        lvRegionConfigPtrTbl[lvRank] = (&regionConfigTbl2[i]);
        lvDscNum += regionConfigTbl2[i].descNum;
    }

    // assign descriptor region index and linking RAM start index
    for (i = 0; i < lvRegionNum; i++)
    {
        lvRegionConfigPtrTbl[i]->memRegion = (Qmss_MemRegion) (MY_DSP_EM_FIRST_MEMORY_REGION_IDX + i);
        lvRegionConfigPtrTbl[i]->startIndex = lvLinkIdx;
        lvLinkIdx += lvRegionConfigPtrTbl[i]->descNum;
    }

    // update input descriptor regions
    for (i = 0; i < regionNum1; i++)
    {
        regionConfigTbl1[i].memRegion = regionConfigTbl2[Emti_REGION_NUM + i].memRegion;
        regionConfigTbl1[i].startIndex = regionConfigTbl2[Emti_REGION_NUM + i].startIndex;
    }

    return lvRegionNum;
}

em_status_t
Emti_initQmss (
    uint32_t regionNum,
    Qmss_MemRegInfo * regionConfigTbl,
    uint32_t pdspNum,
    Qmss_PdspCfg * pdspConfigTbl,
    uint8_t * extLinkTbl,
    Cppi_Handle * cppiHdlPtr
)
{
    uint32_t lvDscNum = 0;
    Qmss_InitCfg lvQmssCfg;
    Cppi_CpDmaInitCfg lvCppiCfg;
    int i, j;

    for (i = 0; i < regionNum; i++)
    {
        lvDscNum += regionConfigTbl[i].descNum;
    }

    /*
     * Set up the linking RAM configuration. Use the internal Linking RAM. 
     */
    lvQmssCfg.linkingRAM0Base = 0;  // Use linking RAM of Navigator SS
#ifdef DSP_EXAMPLE_STANDALONE
    lvQmssCfg.linkingRAM0Size = 0;  // Use all linking RAM of Navigator SS
    lvQmssCfg.linkingRAM1Base = (uint32_t) extLinkTbl;
#else
    /*
     * Linux already configured QMSS 
     * Retrieving the value
     */
    {
        CSL_Qm_configRegs *qm_config_regs_ptr;
#ifdef TI_EM_C6634
        qm_config_regs_ptr = (CSL_Qm_configRegs *) CSL_QMSS_CFG_QM_1_CFG_REGS;
#else
        qm_config_regs_ptr = (CSL_Qm_configRegs *) CSL_QM_SS_CFG_CONFIG_STARVATION_COUNTER_REGS;
#endif
        lvQmssCfg.linkingRAM0Size = qm_config_regs_ptr->LINKING_RAM_REGION_0_SIZE_REG;
        lvQmssCfg.linkingRAM1Base = qm_config_regs_ptr->LINKING_RAM_REGION_1_BASE_ADDRESS_REG;
    }
#endif
    lvQmssCfg.maxDescNum = MY_DSP_EM_FIRST_DESC_IDX_IN_LINKING_RAM + lvDscNum;
#ifdef TI_EM_C6634
    lvQmssCfg.mode = Qmss_Mode_JOINT_LOADBALANCED;
#endif

    // Set up the PDSP configuration
    for (i = 0; i < pdspNum; i++)
    {
        lvQmssCfg.pdspFirmware[i].pdspId = pdspConfigTbl[i].pdspId;
        lvQmssCfg.pdspFirmware[i].firmware = pdspConfigTbl[i].firmware;
        lvQmssCfg.pdspFirmware[i].size = pdspConfigTbl[i].size;
    }
    for (i = pdspNum; i < QMSS_MAX_PDSP; i++)
    {
        lvQmssCfg.pdspFirmware[i].pdspId = (Qmss_PdspId) i;
        lvQmssCfg.pdspFirmware[i].firmware = NULL;
        lvQmssCfg.pdspFirmware[i].size = 0;
    }

    /*
     * Initialize the QMSS 
     */
    if (Qmss_init (&lvQmssCfg, (Qmss_GlobalConfigParams *) & qmssGblCfgParams) != QMSS_SOK)
        return EM_ERR;

    /*
     * Start the QMSS. 
     */
    if (Qmss_start () != QMSS_SOK)
        return EM_ERR;

    /*
     * Flush all the queues 
     */
#ifdef DSP_EXAMPLE_STANDALONE
    for (i = 0; i < qmssGblCfgParams.maxQue; i++)
        Qmss_queueEmpty (i);
#else
    /*
     * Flush the queues used by my instance
     */
    for (i = 0; i < TI_EM_DMA_QUEUE_NUM; i++)
        Qmss_queueEmpty (i + MY_EM_DMA_QUEUE_BASE_IDX);
    for (i = 0; i < TI_EM_HW_QUEUE_NUM; i++)
        Qmss_queueEmpty (i + MY_EM_HW_QUEUE_BASE_IDX);
    for (i = 0; i < 10; i++)
        Qmss_queueEmpty (i + MY_EM_CD_PRIVATE_FREE_QUEUE_IDX);
#endif

    // Insert descriptor regions.
    for (i = 0; i < regionNum; i++)
    {
        for (j = 0; j < regionNum; j++)
        {
            if (regionConfigTbl[j].memRegion == (Qmss_MemRegion) (MY_DSP_EM_FIRST_MEMORY_REGION_IDX + i))
            {
                if (Qmss_insertMemoryRegion (&regionConfigTbl[j]) < 0)
                    return EM_ERR;
            }
        }
    }

    /*
     * Initialize CPPI CPDMA 
     */
#ifdef TI_EM_C6634
    if (Cppi_init (&cppiGblCfgParams) != CPPI_SOK)
        return EM_ERR;
#else
    if (Cppi_init (cppiGblCfgParams) != CPPI_SOK)
        return EM_ERR;
#endif

    /*
     * Open cppi 
     */
    lvCppiCfg.dmaNum = Cppi_CpDma_QMSS_CPDMA;
    lvCppiCfg.writeFifoDepth = 3;   // 
    lvCppiCfg.timeoutCount = 0; //

#ifdef TI_EM_C6634
    lvCppiCfg.qm0BaseAddress = CSL_QMSS_DATA_QM1_QUEUE_MANAGEMENT_REGS + 0 * 4096;
    lvCppiCfg.qm1BaseAddress = CSL_QMSS_DATA_QM1_QUEUE_MANAGEMENT_REGS + 16 * 4096;
    lvCppiCfg.qm2BaseAddress = CSL_QMSS_DATA_QM1_QUEUE_MANAGEMENT_REGS + 32 * 4096;
    lvCppiCfg.qm3BaseAddress = CSL_QMSS_DATA_QM1_QUEUE_MANAGEMENT_REGS + 48 * 4096;
#else
    lvCppiCfg.qm0BaseAddress = CSL_QM_SS_DATA_QM_QUEUE_DEQUEUE_REGS + 0 * 4096;
    lvCppiCfg.qm1BaseAddress = CSL_QM_SS_DATA_QM_QUEUE_DEQUEUE_REGS + 16 * 4096;
    lvCppiCfg.qm2BaseAddress = CSL_QM_SS_DATA_QM_QUEUE_DEQUEUE_REGS + 32 * 4096;
    lvCppiCfg.qm3BaseAddress = CSL_QM_SS_DATA_QM_QUEUE_DEQUEUE_REGS + 48 * 4096;
#endif /* TI_EM_C6634 */

    /*
     * Open QMSS CPDMA 
     */
    *cppiHdlPtr = Cppi_open (&lvCppiCfg);

    /*
     * CPPI and QMSS Infrastructure is operational. 
     */
    return EM_OK;
}

em_status_t
Emti_initPrivateEvents (
    Qmss_MemRegInfo * regionConfigTbl2, ti_em_hw_config2_t hw_config2
)
{
    Qmss_QueueHnd lvPrivateFreeQueueHandle;
    uint8_t lvIsAllocated;
    uint32_t lvEventDscAdd;
    uint32_t lvIndex;

    /*
     * Open the queues 
     */
    lvPrivateFreeQueueHandle =
        Qmss_queueOpen (Qmss_QueueType_GENERAL_PURPOSE_QUEUE, hw_config2.cd_region_queue_idx, &lvIsAllocated);
    if ((lvPrivateFreeQueueHandle < 0) || (lvIsAllocated != 1))
        return EM_ERR_ALLOC_FAILED;

    /*
     * assume private region is at base address of the pdsp data ram, 
     * first descriptor is reserved for cd queue identification 
     */
    lvEventDscAdd = (uint32_t) (regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descBase) + TI_EM_PRIVATE_EVENT_DSC_SIZE;
    /*
     * push private descriptors in private free queue 
     */
    for (lvIndex = 0; lvIndex < (TI_EM_CD_PRIVATE_EVENT_NUM - 1); lvIndex++)
    {
        Qmss_queuePushDesc (lvPrivateFreeQueueHandle, (void *) lvEventDscAdd);
        lvEventDscAdd += TI_EM_PRIVATE_EVENT_DSC_SIZE;
    }

    lvPrivateFreeQueueHandle =
        Qmss_queueOpen (Qmss_QueueType_GENERAL_PURPOSE_QUEUE, hw_config2.ap_region_queue_idx, &lvIsAllocated);
    if ((lvPrivateFreeQueueHandle < 0) || (lvIsAllocated != 1))
        return EM_ERR_ALLOC_FAILED;

    lvEventDscAdd = (uint32_t) (regionConfigTbl2[Emti_PRIVATE1_REGION_IDX].descBase);
    /*
     * push private descriptors in private free queue 
     */
    for (lvIndex = 0; lvIndex < TI_EM_AP_PRIVATE_EVENT_NUM; lvIndex++)
    {
        Qmss_queuePushDesc (lvPrivateFreeQueueHandle, (void *) lvEventDscAdd);
        lvEventDscAdd += TI_EM_PRIVATE_EVENT_DSC_SIZE;
    }

    return EM_OK;
}

em_status_t
Emti_initEvents (
    Qmss_MemRegInfo * regionConfigTbl,
    uint32_t poolNum,
    ti_em_pool_config2_t * poolConfig2Tbl
)
{
    uint8_t *lvDscPtr;
    uint32_t lvDscSize;
    uint32_t lvEventNumTotal = 0;
    int i;

    lvDscPtr = (uint8_t *) regionConfigTbl[Emti_PUBLIC_REGION_IDX].descBase;
    lvDscSize = regionConfigTbl[Emti_PUBLIC_REGION_IDX].descSize;
    lvDscPtr += lvDscSize;

    for (i = 0; i < poolNum; i++)
    {
        uint8_t *lvBufPtr;
        uint32_t lvBufSize;
        uint32_t lvEventNum;
        uint32_t lvQueueIdx;
        int j;

        lvBufPtr = poolConfig2Tbl[i].buffer_ptr;
        lvBufSize = poolConfig2Tbl[i].pool_config.buf_size;
        lvEventNum = poolConfig2Tbl[i].event_num;
        lvQueueIdx = poolConfig2Tbl[i].pool_config.free_queue_idx;
        lvEventNumTotal += lvEventNum;

        if (lvEventNumTotal > TI_EM_EVENT_NUM2)
            return EM_ERR;
        if (ti_em_hw_queue_open (lvQueueIdx) != EM_OK)
            return EM_ERR;

        for (j = 0; j < lvEventNum; j++)
        {
            Cppi_DescTag lvTag;

            ti_em_packet_set_default ((ti_em_packet_t *) lvDscPtr);

            ti_em_packet_set_pool_info ((ti_em_packet_t *) (lvDscPtr), (poolConfig2Tbl[i].pool_config));

            if (poolConfig2Tbl[i].buffer_ptr != NULL)
            {
                ti_em_buffer_config_t lvBufferConfig;

                lvBufferConfig.buffer_ptr = (void *) (lvBufPtr);
                lvBufferConfig.buffer_size = (uint32_t) (lvBufSize);
                lvBufferConfig.orig_buffer_ptr = (void *) (lvBufPtr);
                lvBufferConfig.orig_buffer_size = (uint32_t) (lvBufSize);

                lvBufferConfig.coh_mode = (ti_em_coh_mode_t) (poolConfig2Tbl[i].coh_mode);
                lvBufferConfig.free_func = (ti_em_free_func_t) NULL;
                lvBufferConfig.free_push_policy = (ti_em_push_policy_t) TI_EM_PUSH_POLICY_TAIL;
                ti_em_packet_set_buffer_info ((ti_em_packet_t *) lvDscPtr, lvBufferConfig);
            }

            lvTag.srcTagHi = 0x00;
            lvTag.srcTagLo = 0xFF;
            lvTag.destTagHi = 0x00;
            lvTag.destTagLo = 0x00;
            Cppi_setTag (Cppi_DescType_HOST, (Cppi_Desc *) (lvDscPtr), &lvTag);

            __mfence ();

            Qmss_queuePushDesc (lvQueueIdx, (void *) lvDscPtr);

            lvBufPtr = lvBufPtr + lvBufSize;
            lvDscPtr = lvDscPtr + lvDscSize;
        }
    }

    return EM_OK;
}

void
Emti_initHwConfig (
    ti_em_hw_config2_t hwConfig2,
    Qmss_MemRegInfo * regionConfigTbl,
    Cppi_Handle cppiHdl,
    uint32_t poolNum,
    ti_em_pool_config2_t * poolConfig2Tbl,
    ti_em_config_t * hwConfigPtr,
    Qmss_PdspId emPdspIdx
)
{
    int i;

    hwConfigPtr->my_device_idx = 0;
    hwConfigPtr->my_process_idx = 0;
    hwConfigPtr->chain_config_ptr = NULL;
    hwConfigPtr->preload_config_ptr = NULL;
    hwConfigPtr->pdsp_num = 1;
    hwConfigPtr->pdsp_idx_tbl[0] = emPdspIdx;
    hwConfigPtr->ap_region_queue_idx = hwConfig2.ap_region_queue_idx;
    hwConfigPtr->cd_region_queue_idx = hwConfig2.cd_region_queue_idx;
    hwConfigPtr->hw_queue_base_idx = hwConfig2.hw_queue_base_idx;
    hwConfigPtr->dma_queue_base_idx = hwConfig2.dma_queue_base_idx;
    hwConfigPtr->pool_num = poolNum;
    for (i = 0; i < poolNum; i++)
    {
        hwConfigPtr->pool_config_tbl[i].free_queue_idx = poolConfig2Tbl[i].pool_config.free_queue_idx;
        hwConfigPtr->pool_config_tbl[i].buf_mode = poolConfig2Tbl[i].pool_config.buf_mode;
        hwConfigPtr->pool_config_tbl[i].dsc_wsize = poolConfig2Tbl[i].pool_config.dsc_wsize;
    }
    hwConfigPtr->hw_sem_idx = hwConfig2.hw_sem_idx;
    hwConfigPtr->dma_idx = (uint32_t) cppiHdl;
}

em_status_t
ti_em_init_global2 (
    uint32_t region_num,
    Qmss_MemRegInfo * region_config_tbl,
    uint32_t pdsp_num,
    Qmss_PdspCfg * pdsp_config_tbl,
    uint32_t pool_num,
    ti_em_pool_config2_t * pool_config2_tbl,
    ti_em_hw_config2_t hw_config2
)
{
    Qmss_MemRegInfo lvRegionCfgTbl[Emti_REGION_NUM + 
                                   TOMP_NUM_QMSS_MEMORY_REGIONS];
    int32_t lvRegionNum;
    Qmss_PdspCfg lvPdspCfgTbl[Emti_PDSP_NUM];
    int32_t lvPdspNum;
    ti_em_config_t lvHwConfig;
    Cppi_Handle lvCppiHdl;
    Qmss_PdspId lvEmPdspIdx;

    if (pool_num > TI_EM_POOL_NUM)
        return EM_ERR;

    // Build table of PDSP configurations
    lvPdspNum = Emti_initPdspConfig (pdsp_num, pdsp_config_tbl, lvPdspCfgTbl, &lvEmPdspIdx);
    if (lvPdspNum < 0)
        return EM_ERR;

    // Build table of region configurations
    lvRegionNum = Emti_initRegionConfig (region_num, region_config_tbl, lvRegionCfgTbl);
    if (lvRegionNum < 0)
        return EM_ERR;

    // Initialize QMSS
    if (Emti_initQmss (lvRegionNum, lvRegionCfgTbl, lvPdspNum, lvPdspCfgTbl, NULL, &lvCppiHdl) != EM_OK)
        return EM_ERR;

    // initialize private free queue
    if (Emti_initPrivateEvents(lvRegionCfgTbl, hw_config2) != EM_OK)  return EM_ERR;

    // Build table of pool configurations and initialize events
    if (Emti_initEvents (lvRegionCfgTbl, pool_num, pool_config2_tbl) != EM_OK)
        return EM_ERR;

    // Build table of preload pool configurations and initialize local events

    // Set EM configuration and initialize EM
    Emti_initHwConfig (hw_config2, lvRegionCfgTbl, lvCppiHdl, pool_num, pool_config2_tbl,
                       &lvHwConfig, lvEmPdspIdx);
    if (ti_em_init_global (&lvHwConfig) != EM_OK)
        return EM_ERR;

/*
 * The process connection has to be done first
 */
#if 0
// This was for communicating to ARM side??? in example 2???
    if (ti_em_connect (MY_EM_DEVICE_CD_QUEUE_IDX) != EM_OK)
        return EM_ERR;

    if (ti_em_process_add_route (MY_LINUX_EM_FLOW_IDX, MY_LINUX_EM_PROCESS_IDX) != EM_OK)
        return EM_ERR;
#endif

    return EM_OK;
}

em_status_t
ti_em_init_local2 (
    void
)
{
    if (Qmss_start () != QMSS_SOK)
        return EM_ERR;

    if (ti_em_init_local () != EM_OK)
        return EM_ERR;

    return EM_OK;
}
